from functools import partial
from typing import Literal
import concurrent.futures
import asyncio
from collections.abc import AsyncGenerator
from pathlib import Path

from sqlalchemy import URL
from sqlalchemy.ext.asyncio import create_async_engine, AsyncEngine
from sqlalchemy.util import immutabledict
from sqlmodel.ext.asyncio.session import AsyncSession
import pandas as pd

from knoc.settings import settings

engine_async_url: URL = URL(
    drivername="postgresql+psycopg",
    username=settings.db.username,
    password=settings.db.password,
    host=settings.db.host,
    port=settings.db.port,
    database=settings.db.database,
    query=immutabledict(),
)

engine_sync_url: URL = URL(
    drivername="postgresql",
    username=settings.db.username,
    password=settings.db.password,
    host=settings.db.host,
    port=settings.db.port,
    database=settings.db.database,
    query=immutabledict(),
)


engine: AsyncEngine = create_async_engine(engine_async_url, echo=settings.db.echo)


async def get_session() -> AsyncGenerator[AsyncSession]:
    async with AsyncSession(engine) as session:
        yield session


async def aio_df_to_sql(df: pd.DataFrame, tablename: str, *args, **kwargs):
    """Utility function to save a pandas df to a pg table in async.
    The extra arguments (args and kwargs) are passed directly to df.to_sql"""
    loop = asyncio.get_running_loop()
    with concurrent.futures.ProcessPoolExecutor() as pool:
        await loop.run_in_executor(
            pool,
            partial(
                df.to_sql,
                tablename,
                engine_sync_url.render_as_string(hide_password=False),
                *args,
                **kwargs,
            ),
        )


async def aio_sql_to_df(
    tablename: str, schema: str = "public", *args, **kwargs
) -> pd.DataFrame:
    """Utility function to read a pq table to a pandas df table in async.
    The extra arguments (args and kwargs) are passed directly to df.read_sql"""
    loop = asyncio.get_running_loop()
    with concurrent.futures.ProcessPoolExecutor() as pool:
        df = await loop.run_in_executor(
            pool,
            partial(
                pd.read_sql,
                f'select * from "{schema}"."{tablename}"',
                engine_sync_url.render_as_string(hide_password=False),
                *args,
                **kwargs,
            ),
        )
    return df


def csv_to_table(
    csv_file: Path,
    tablename: str,
    schema: str = "public",
    if_exists: Literal["fail", "replace", "append"] = "fail",
    delimiter: str = ";",
    quotechar: str = '"',
) -> pd.DataFrame:
    """Import a csv export to a table"""
    df = pd.read_csv(csv_file, delimiter=delimiter, quotechar=quotechar)
    df.to_sql(
        tablename,
        schema=schema,
        if_exists=if_exists,
        con=engine_sync_url.render_as_string(hide_password=False),
    )
    return df
