#!/usr/bin/env python
"""Knoc - KNetwork Orchectration Center"""

import logging
from enum import Enum
from importlib.metadata import entry_points

import typer
import uvicorn
from uvicorn.config import LOG_LEVELS

from knoc.worker import app as app_worker
from knoc.cli.workflow import app as app_workflow

app = typer.Typer(
    help=__doc__,
    no_args_is_help=True,
)
app.add_typer(app_worker, name="worker", help="Manage workers")
app.add_typer(app_workflow, name="workflow", help="Manage workflows")

FORMAT = "%(name)s: %(message)s"
logging.basicConfig(
    level=logging.INFO,
    format=FORMAT,
    datefmt="[%y-%m-%d %X]",
)

logger = logging.getLogger("knoc")

LogLevels = Enum("LogLevel", {k: k for k in LOG_LEVELS})
# Nice, but makes auto-completion sluggish
# Workflows = Enum("Workflows", {k.name: k.name for k in all_wfs})


def version_callback(value):
    if value:
        from knoc import __version__

        typer.echo(__version__)
        raise typer.Exit()


@app.command()
def serve(
    host: str = "localhost",
    port: int = 8000,
    log_level: LogLevels = "info",  # type: ignore
):
    """Start the knoc engine"""
    logging.root.setLevel(LOG_LEVELS[log_level.name])
    uvicorn.run("knoc.server:app", host=host, port=port, log_level=log_level.name)


@app.callback()
def common(
    version: bool = typer.Option(
        None,
        "--version",
        "-v",
        help="Print version and exit",
        callback=version_callback,
    ),
):
    pass


def get_plugins():
    for entry_point in entry_points().select(group="knoc.cli"):
        try:
            sub_app = entry_point.load()
        except Exception as err:
            logger.warning(f"Cannot load CLI module {entry_point.name} - see below")
            logger.exception(err)
            continue
        if not isinstance(sub_app, typer.Typer):
            logger.warning(f"CLI module {entry_point.name} is not a Typer")
            continue
        app.add_typer(
            sub_app, name=entry_point.name, help=f"Help for {entry_point.name}"
        )


get_plugins()


if __name__ == "__main__":
    app()
